Attribute VB_Name = "mAccess"
Option Explicit

Private Enum ACCESS_MODE
    NOT_USED_ACCESS = 0
    GRANT_ACCESS = 1
    SET_ACCESS = 2
    DENY_ACCESS = 3
    REVOKE_ACCESS = 4
    SET_AUDIT_SUCCESS = 5
    SET_AUDIT_FAILURE = 6
End Enum

Private Enum MULTIPLE_TRUSTEE_OPERATION
    NO_MULTIPLE_TRUSTEE = 0
    TRUSTEE_IS_IMPERSONATE = 1
End Enum

Private Enum TRUSTEE_FORM
    TRUSTEE_IS_SID = 0
    TRUSTEE_IS_NAME = 1
    TRUSTEE_BAD_FORM = 2
    TRUSTEE_IS_OBJECTS_AND_SID = 3
    TRUSTEE_IS_OBJECTS_AND_NAME = 4
End Enum

Private Enum TRUSTEE_TYPE
    TRUSTEE_IS_UNKNOWN = 0
    TRUSTEE_IS_USER = 1
    TRUSTEE_IS_GROUP = 2
    TRUSTEE_IS_DOMAIN = 3
    TRUSTEE_IS_ALIAS = 4
    TRUSTEE_IS_WELL_KNOWN_GROUP = 5
    TRUSTEE_IS_DELETED = 5
    TRUSTEE_IS_INVALID = 6
    TRUSTEE_IS_COMPUTER = 7
End Enum

Private Type SID_IDENTIFIER_AUTHORITY
    Value(6) As Byte
End Type

Private Type TRUSTEE
    pMultipleTrustee As Long
    MultipleTrusteeOperation As MULTIPLE_TRUSTEE_OPERATION
    TrusteeForm As TRUSTEE_FORM
    TrusteeType As TRUSTEE_TYPE
    ptstrName As Long
End Type

Private Type EXPLICIT_ACCESS
    grfAccessPermissions As Long
    grfAccessMode As ACCESS_MODE
    grfInheritance As Long
    TRUSTEE As TRUSTEE
End Type
    
Public Type SECURITY_ATTRIBUTES
    nLength As Long
    lpSecurityDescriptor As Long
    bInheritHandle As Long
End Type

Private Type ACL
   AclRevision As Byte
   Sbz1 As Byte
   AclSize As Integer
   AceCount As Integer
   Sbz2 As Integer
End Type

Private Type SECURITY_DESCRIPTOR
   Revision As Byte
   Sbz1 As Byte
   Control As Long
   Owner As Long
   Group As Long
   sACL As ACL
   Dacl As ACL
End Type

Private Declare Function AllocateAndInitializeSid Lib "advapi32.dll" (pIdentifierAuthority As SID_IDENTIFIER_AUTHORITY, ByVal nSubAuthorityCount As Byte, ByVal nSubAuthority0 As Long, ByVal nSubAuthority1 As Long, ByVal nSubAuthority2 As Long, ByVal nSubAuthority3 As Long, ByVal nSubAuthority4 As Long, ByVal nSubAuthority5 As Long, ByVal nSubAuthority6 As Long, ByVal nSubAuthority7 As Long, lpPSid As Long) As Long
Private Declare Sub FreeSid Lib "advapi32.dll" (pSid As Any)
Private Declare Function GetLastError Lib "kernel32" () As Long
Private Declare Function InitializeSecurityDescriptor Lib "advapi32.dll" (pSecurityDescriptor As SECURITY_DESCRIPTOR, ByVal dwRevision As Long) As Long
Private Declare Function LocalAlloc Lib "kernel32" (ByVal wFlags As Long, ByVal wBytes As Long) As Long
Private Declare Function LocalFree Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function SetEntriesInAcl Lib "advapi32.dll" Alias "SetEntriesInAclA" (ByVal cCountOfExplicitEntries As Long, ByRef pListOfExplicitEntries() As EXPLICIT_ACCESS, ByVal OldAcl As Long, ByRef NewAcl As Long) As Long
Private Declare Function SetSecurityDescriptorDacl Lib "advapi32.dll" (pSecurityDescriptor As SECURITY_DESCRIPTOR, ByVal bDaclPresent As Long, pDacl As Long, ByVal bDaclDefaulted As Long) As Long
Private Declare Sub ZeroMemory Lib "kernel32.dll" Alias "RtlZeroMemory" (Destination As Any, ByVal Length As Long)

Private Const SECURITY_WORLD_RID = &H0
Private Const SECURITY_WORLD_SID_AUTHORITY = 1
Private Const SECURITY_DESCRIPTOR_REVISION = 1
Private Const SECURITY_NT_AUTHORITY          As Long = &H5
Private Const SECURITY_BUILTIN_DOMAIN_RID    As Long = &H20&
Private Const STANDARD_RIGHTS_ALL = &H1F0000
Private Const READ_CONTROL = &H20000
Private Const STANDARD_RIGHTS_READ = (READ_CONTROL)
Private Const KEY_QUERY_VALUE = &H1
Private Const KEY_SET_VALUE = &H2
Private Const KEY_CREATE_SUB_KEY = &H4
Private Const KEY_ENUMERATE_SUB_KEYS = &H8
Private Const KEY_NOTIFY = &H10
Private Const KEY_CREATE_LINK = &H20
Private Const SYNCHRONIZE = &H100000
Private Const KEY_ALL_ACCESS = ((STANDARD_RIGHTS_ALL Or _
                                 KEY_QUERY_VALUE Or _
                                 KEY_SET_VALUE Or _
                                 KEY_CREATE_SUB_KEY Or _
                                 KEY_ENUMERATE_SUB_KEYS Or _
                                 KEY_NOTIFY Or _
                                 KEY_CREATE_LINK) And (Not SYNCHRONIZE))
Private Const KEY_READ = ((STANDARD_RIGHTS_READ Or _
                           KEY_QUERY_VALUE Or _
                           KEY_ENUMERATE_SUB_KEYS Or _
                           KEY_NOTIFY) And (Not SYNCHRONIZE))
Private Const NO_INHERITANCE = &H0
Private Const DOMAIN_ALIAS_RID_ADMINS = &H220
Private Const LMEM_FIXED = &H0&
Private Const LMEM_ZEROINIT = &H40&
Private Const LPTR = (LMEM_FIXED Or LMEM_ZEROINIT)

Public Function SetAccess() As SECURITY_ATTRIBUTES
    Dim dwRes As Long
    Dim pEveryoneSID As Long, pAdminSID As Long
    Dim pACL As Long
    Dim pSD As SECURITY_DESCRIPTOR
    Dim ea(1) As EXPLICIT_ACCESS
    Dim SIDAuthWorld As SID_IDENTIFIER_AUTHORITY
    Dim SIDAuthNT As SID_IDENTIFIER_AUTHORITY
    Dim sa As SECURITY_ATTRIBUTES

    '    SID   ""
    SIDAuthNT.Value(5) = SECURITY_NT_AUTHORITY
    SIDAuthWorld.Value(5) = SECURITY_WORLD_SID_AUTHORITY
    If AllocateAndInitializeSid(SIDAuthWorld, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0, pEveryoneSID) = ERROR_SUCCESS Then
        Err.Raise 5, App.EXEName, "AllocateAndInitializeSid Error " & GetLastError
        GoTo Cleanup
    End If

    '   EXPLICIT_ACCESS  ACE.
    ' ACE       .
    Call ZeroMemory(ea(0), Len(ea(0)))
    With ea(0)
        .grfAccessPermissions = KEY_READ
        .grfAccessMode = SET_ACCESS
        .grfInheritance = NO_INHERITANCE
        With .TRUSTEE
            .TrusteeForm = TRUSTEE_IS_SID
            .TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP
            .ptstrName = VarPtr(pEveryoneSID)
        End With
    End With
    
    '  SID   BUILTIN\Administrators
    If AllocateAndInitializeSid(SIDAuthNT, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, pAdminSID) = ERROR_SUCCESS Then
        Err.Raise 5, App.EXEName, "AllocateAndInitializeSid Error" & GetLastError()
        GoTo Cleanup
    End If

    '   EXPLICIT_ACCESS  ACE.
    ' ACE        
    With ea(1)
        .grfAccessPermissions = KEY_ALL_ACCESS
        .grfAccessMode = SET_ACCESS
        .grfInheritance = NO_INHERITANCE
        With .TRUSTEE
            .TrusteeForm = TRUSTEE_IS_SID
            .TrusteeType = TRUSTEE_IS_GROUP
            .ptstrName = VarPtr(pAdminSID)
        End With
    End With

    '   ,    ACEs.
    dwRes = SetEntriesInAcl(2, ea, 0&, pACL)
    If dwRes <> ERROR_SUCCESS Then
        Err.Raise 5, App.EXEName, "SetEntriesInAcl Error " & dwRes
        GoTo Cleanup
    End If

    '   
    
    pSD.Control = LocalAlloc(LPTR, Len(pSD))
    If pSD.Control = 0 Then
        Err.Raise 5, App.EXEName, "LocalAlloc Error " & GetLastError()
        GoTo Cleanup
    End If
    If InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION) = ERROR_SUCCESS Then
        Err.Raise 5, App.EXEName, "InitializeSecurityDescriptor Error " & GetLastError()
        GoTo Cleanup
    End If
 
    '  ACL   
    If SetSecurityDescriptorDacl(pSD, True, pACL, False) = ERROR_SUCCESS Then
        Err.Raise 5, App.EXEName, "SetSecurityDescriptorDacl Error " & GetLastError()
        GoTo Cleanup
    End If
    
    '    
    With sa
        .nLength = Len(sa)          '    
        .lpSecurityDescriptor = pSD.Control    '     
        .bInheritHandle = False     '    ,
                                    '  Windows 9x 
    End With
    SetAccess = sa
Cleanup:
    If CBool(pEveryoneSID) Then FreeSid pEveryoneSID
    If CBool(pAdminSID) Then FreeSid pAdminSID
    If CBool(pACL) Then LocalFree pACL
    If CBool(pSD.Control) Then LocalFree pSD.Control
End Function
